# Copyright 2019 John Reese
# Licensed under the MIT license

from pathlib import Path

import click

from .core import Converter, setup_logger


@click.command(help="convert reStructuredText annotations to python type stubs")
@click.option("--debug", "-D", is_flag=True, default=False, help="enable debug logging")
@click.argument(
    "source_dir", type=click.Path(exists=True, file_okay=False, resolve_path=False)
)
@click.argument(
    "dest_dir",
    type=click.Path(exists=False, file_okay=False, writable=True, resolve_path=False),
)
def main(debug: bool = False, source_dir: str = ".", dest_dir: str = "."):
    setup_logger(debug)
    source = Path(source_dir)
    dest = Path(dest_dir)
    Converter(source, dest).gen_stubs()


if __name__ == "__main__":
    main()



# Copyright 2019 John Reese
# Licensed under the MIT license

from pathlib import Path
from typing import List, Sequence, Set

from attr import dataclass


@dataclass
class Config:
    # regex to match lines - unmatched lines will not be converted
    line_prefix: str = r"\s*//\|"
    ignore_directives: Set[str] = {"warning", "note", "currentmodule"}

    module_template: str = '\n"""\n{name}\n"""\n'
    class_template: str = (
        "\n# {source}:{lineno}\nclass {name}:\n    def __init__(self, {args}): ..."
    )
    method_template: str = "    def {name}(self, {args}) -> {ret_type}: ..."
    attribute_template: str = "    {name}: {attr_type} = ..."
    function_template: str = (
        "\n# {source}:{lineno}\ndef {name}({args}) -> {ret_type}: ..."
    )
    arg_template: str = "{name}: {arg_type}{equal}{value}"
    import_template: str = "import {name}"
    from_template: str = "from {module} import {name}"


@dataclass
class Line:
    source: Path
    lineno: int
    kind: str
    extra: Sequence[str]


Lines = List[Line]



from setuptools import setup

with open("README.md") as f:
    readme = f.read()

with open("rst2pyi/__init__.py") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.split('"')[1]

with open("requirements.txt") as f:
    requirements = [line.strip() for line in f]

setup(
    name="rst2pyi",
    description="convert reStructuredText annotations to PEP 484 type stubs",
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version,
    author="John Reese",
    author_email="john@noswap.com",
    url="https://github.com/jreese/rst2pyi",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
    ],
    license="MIT",
    packages=["rst2pyi", "rst2pyi.tests"],
    setup_requires=["setuptools>=38.6.0"],
    install_requires=requirements,
    tests_require=[],
    test_suite="rst2pyi.tests",
    entry_points={"console_scripts": ["rst2pyi = rst2pyi.main:main"]},
)



# Copyright 2019 John Reese
# Licensed under the MIT license

import builtins
import logging
import re
import typing
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Set

from attr import astuple

from .types import Config, Line, Lines

log = logging.getLogger(__name__)
directive_re = re.compile(r"\s*..\s+(\w+)::\s+([^\n]+)")
callable_re = re.compile(r"\s*(\w+)(?:\(([^\)]*)\))?")
param_re = re.compile(r"\s*:(\w+)\s+(.+)\s+(\w+):")


def setup_logger(debug: bool = False):
    logging.addLevelName(logging.DEBUG, "DBG")
    logging.addLevelName(logging.INFO, "INF")
    logging.addLevelName(logging.WARNING, "WRN")
    logging.addLevelName(logging.ERROR, "ERR")
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format="%(levelname)s %(message)s",
    )


class Converter:
    def __init__(self, source_dir: Path, dest_dir: Path, **kwargs: Any):
        self.source_dir = source_dir
        self.dest_dir = dest_dir
        self.config = Config(**kwargs)

        self.prefix_re = re.compile(self.config.line_prefix)

    def parse_file(self, source: Path) -> Lines:
        lines = []
        prefix_re = self.prefix_re
        with open(source, "r") as f:
            for lineno, content in enumerate(f, start=1):
                match = prefix_re.match(content)
                if match:
                    content = content[match.end() + 1 :]
                    dmatch = directive_re.match(content)
                    pmatch = param_re.match(content)

                    if dmatch:
                        kind, *extra = dmatch.groups()
                        lines.append(Line(source, lineno, kind, extra))
                    elif pmatch:
                        kind, *extra = pmatch.groups()
                        lines.append(Line(source, lineno, kind, extra))

        return lines

    def parse_dir(self, source_dir: Path) -> Dict[Path, Lines]:
        source_dir = source_dir.expanduser()
        docs: Dict[Path, Lines] = {}

        for path in source_dir.iterdir():
            if path.is_dir():
                docs.update(self.parse_dir(path))
            elif path.is_file():
                lines = self.parse_file(path)
                if lines:
                    docs[path] = lines

        return docs

    def organize(self, contents: Dict[Path, Lines]) -> Dict[str, Lines]:
        modules: Dict[str, Lines] = defaultdict(list)
        for _path, lines in contents.items():
            module: str = "builtins"
            for line in lines:
                if line.kind in ("module", "currentmodule"):
                    module = line.extra[0]
                if line.kind not in self.config.ignore_directives:
                    modules[module].append(line)
        for name in modules:
            modules[name] = list(
                sorted(modules[name], key=lambda l: 1 if l.kind == "module" else 2)
            )
        return modules

    def render(self, line: Line, kind: str = None, **kwargs: str) -> str:
        if kind is None:
            kind = line.kind
        tpl = getattr(self.config, f"{kind}_template")
        return tpl.format(
            source=line.source,
            lineno=line.lineno,
            kind=kind,
            extra=line.extra,
            **kwargs,
        )

    def gen_stub(self, dest: Path, lines: Lines) -> None:
        log.debug("generating stub %s", dest)
        config = self.config
        content: List[str] = []
        type_names: Set[str] = set()
        count = len(lines)
        idx = 0
        while idx < count:
            line = lines[idx]
            path, lineno, kind, extra = astuple(line)
            if kind in ("module",):
                name, *_ = extra
                content.append(self.render(line, name=name))

            elif kind in ("class", "method", "function"):
                call, = extra
                match = callable_re.match(call)
                if not match:
                    idx += 1
                    log.error(
                        "%s:%d: failed to parse %s directive: %s",
                        path,
                        lineno,
                        kind,
                        call,
                    )
                    continue

                name, param_str = match.groups()
                if param_str is None:
                    content.append(
                        self.render(line, name=name, args="", ret_type="Any")
                    )
                    idx += 1
                    continue

                params = [
                    [n.strip(" []"), "Any", v.strip(" []")]
                    for p in param_str.split(",")
                    for n, _, v in [p.partition("=")]
                ]

                while idx + 1 < count and lines[idx + 1].kind == "param":
                    idx += 1
                    p_type, p_name = lines[idx].extra
                    for pidx, (n, _, v) in enumerate(params):
                        if n == p_name:
                            params[pidx][1] = p_type
                            break

                type_names.update(t for _, t, _ in params)
                args = ", ".join(
                    self.render(
                        line,
                        kind="arg",
                        name=n,
                        arg_type=t,
                        equal=" = " if v else "",
                        value=v,
                    )
                    for n, t, v in params
                )
                content.append(self.render(line, name=name, args=args, ret_type="Any"))
            elif kind in ("attribute", "data"):
                name, *_ = extra
                attr_type = "Any"
                type_names.add(attr_type)
                content.append(
                    self.render(line, kind="attribute", name=name, attr_type=attr_type)
                )
            elif kind in ("currentmodule", "note", "warning"):
                pass  # ignore these directives
            else:
                log.warning(
                    "%s:%d: unmatched %s directive: %s", path, lineno, kind, extra
                )
            idx += 1

        content.append("")

        imports: Set[str] = set()
        for name in type_names:
            name, _, _ = name.partition(".")
            if hasattr(builtins, name):
                continue
            elif hasattr(typing, name):
                imports.add(config.from_template.format(module="typing", name=name))
            else:
                imports.add(config.import_template.format(name=name))

        content[:0] = list(
            sorted(imports, key=lambda l: f"z{l}" if l.startswith("from") else l)
        )

        with open(dest, "w") as f:
            f.write("\n".join(content))

    def gen_stubs(self) -> None:
        content = self.parse_dir(self.source_dir)
        modules = self.organize(content)
        for module, data in modules.items():
            self.dest_dir.mkdir(parents=True, exist_ok=True)
            dest = self.dest_dir / f"{module}.pyi"
            self.gen_stub(dest, data)
